/*=========================================

	VArray -- utility code for working with
	Handle-based arrays with variable number
	of elements.

	 2000-2001 Brian 'Moses' Hall
	moses@blugs.com
	www.blugs.com

	Heapsort implementation based on code by
	Michael Schrig and Kyle Hammond.

=========================================*/
#include <MacTypes.h>
#include <MacMemory.h>
#include <MacErrors.h>

/* Header and first byte of actual data. */
typedef struct
{
	UInt32		nItems;		/* Items in array. */
	UInt32		itemSize;	/* Size of each data element. */
	UInt32		refCon;		/* Store anything here. */
	UInt32		ownerCount;
	char		data[1];	/* Array data starts here. */
} VARec, **VARef;


#define __VA_INTERNAL__
#include "VArray.h"


#ifndef __VA_DEBUG__
	#define __VA_DEBUG__	0
#endif

#if __VA_DEBUG__
	#define _AssertStr(x,y)						if (!(x)) DebugStr(y)
	#define _Debug(x)							DebugStr(x)
#else
	#define _AssertStr(x,y)
	#define _Debug(x)
#endif

/* Test one or more bits. */
#define va_Test(field,features)		((((field) & (features)) == (features)) ? true : false)

typedef struct
{
	Size	size;
	UInt32	offset;
} VADataSortParam, *VADataSortParamPtr;


static void		local_VAReheap16( UInt16* arrayP, UInt32 left, UInt32 right,
										VASortProcPtr proc, UInt32 refCon );
static void		local_VAReheap32( UInt32* arrayP, UInt32 left, UInt32 right,
										VASortProcPtr proc, UInt32 refCon );
static void		local_VAReheapOther( Ptr arrayP, UInt32 itemSize, UInt32 left, UInt32 right,
										VASortProcPtr proc, UInt32 refCon );
static void		local_VAHeapsort16( UInt16* arrayP, UInt32 nItems, VASortProcPtr proc,
										UInt32 refCon);
static void		local_VAHeapsort32( UInt32* arrayP, UInt32 nItems, VASortProcPtr proc,
										UInt32 refCon);
static void		local_VAHeapsortOther( Ptr arrayP, UInt32 nItems, UInt32 itemSize,
										VASortProcPtr proc, UInt32 refCon);


static VAComparisonResult local_VAByteSortProc( const Ptr data1P, const Ptr data2P,
												UInt32 inOffset );
static VAComparisonResult local_VASignedByteSortProc( const Ptr data1P, const Ptr data2P,
												UInt32 inOffset );
static VAComparisonResult local_VAHalfWordSortProc( const Ptr data1P, const Ptr data2P,
												UInt32 inOffset );
static VAComparisonResult local_VASignedHalfWordSortProc( const Ptr data1P, const Ptr data2P,
												UInt32 inOffset );
static VAComparisonResult local_VAWordSortProc( const Ptr data1P, const Ptr data2P,
												UInt32 inOffset );
static VAComparisonResult local_VASignedWordSortProc( const Ptr data1P, const Ptr data2P,
													UInt32 inOffset );
static VAComparisonResult local_VADataSortProc( const Ptr data1P, const Ptr data2P,
												VADataSortParam* inSortData );



static VAComparisonResult local_VAByteSearchProc( const Ptr inArrayDataP, UInt8* inSearchData,
												UInt32 inOffset );
static VAComparisonResult local_VASignedByteSearchProc( const Ptr inArrayDataP,
														SInt8* inSearchData,
														UInt32 inOffset );
static VAComparisonResult local_VAHalfWordSearchProc( const Ptr inArrayDataP,
													UInt16* inSearchData,
													UInt32 inOffset );
static VAComparisonResult local_VASignedHalfWordSearchProc( const Ptr inArrayDataP,
															SInt16* inSearchData,
															UInt32 inOffset );
static VAComparisonResult local_VAWordSearchProc( const Ptr inArrayDataP, UInt32* inSearchData,
													UInt32 inOffset );
static VAComparisonResult local_VASignedWordSearchProc( const Ptr inArrayDataP,
														SInt32* inSearchData,
														UInt32 inOffset );
static VAComparisonResult local_VADataSearchProc( const Ptr inArrayDataP,
												UInt8* inSearchData,
												VADataSortParam* inSortData );



#define va_N(a)						((*a)->nItems)
#define va_S(a)						((*a)->itemSize)
#define va_RC(a)					((*a)->refCon)
#define va_OC(a)					((*a)->ownerCount)
#define va_D(a)						(&((*a)->data[0]))			
#define va_NthItemPtr(n,data,size)	(data + ((n - 1) * size))
#define va_NthItemPtr16(n,data)		((SInt16*)data + n - 1)
#define va_NthItemPtr32(n,data)		((SInt32*)data + n - 1)


//===============================
VA_PROC VARef VANew( UInt32 inItemCount, UInt32 inItemSize, Boolean inZero, UInt32 inRefCon )
{
	VARef		array = nil;
	UInt32		dataSize, handleSize;
	
	_AssertStr( inItemSize != 0, "\pError: zero item size." );
	if (inItemSize)
	{
		dataSize = inItemCount * inItemSize;
		handleSize = dataSize + sizeof(VARec) - 1;
		array = (VARef)NewHandle( handleSize );
		if (array)
		{
			va_N(array) = inItemCount;
			va_S(array) = inItemSize;
			va_RC(array) = inRefCon;
			if (inZero && dataSize) VAZeroMemory( va_D(array), dataSize );
			va_OC(array) = 1;
		}
	}
	return array;
}

//===============================
VA_PROC void VADispose( VARef inArray )
{
	if (inArray)
	{
		va_OC(inArray)--;
		if (!va_OC(inArray)) DisposeHandle( (Handle)inArray );
	}
}

//===============================
VA_PROC UInt32 VACount( VARef inArray )
{
	return (*inArray)->nItems;
}

//===============================
VA_PROC UInt32 VAItemSize( VARef inArray )
{
	return (*inArray)->itemSize;
}

//===============================
VA_PROC UInt32 VARefCon( VARef inArray )
{
	return (*inArray)->refCon;
}

//===============================
VA_PROC void VASetRefCon( UInt32 inRefCon, VARef inArray )
{
	(*inArray)->refCon = inRefCon;
}

//===============================
VA_PROC UInt32 VAOwnerCount( VARef inArray )
{
	return (*inArray)->ownerCount;
}

//===============================
// Returns a pointer to the first element added.
// In case of error, returns nil.
// If inAddHere is greater than existing items, makes sure it becomes the next item:
//	if there are 3 items and inAddHere == 5, we're actually adding at #4.
VA_PROC void* VAAdd( UInt32 inCount, UInt32 inAddHere, Boolean inZero, VARef inArray )
{
	Ptr		sourceP = nil;
	
	if (inAddHere && inCount)
	{
		UInt32		neededSize, nItems, itemSize;
		
		nItems = va_N(inArray);
		itemSize = va_S(inArray);
		neededSize = sizeof(VARec) + ((nItems + inCount) * itemSize) - 1;
		SetHandleSize( (Handle)inArray, neededSize );
		if (!MemError())
		{
			Ptr		data;
			
			if (inAddHere > nItems + 1) inAddHere = nItems + 1;
			data = va_D(inArray);
			sourceP = va_NthItemPtr(inAddHere,data,itemSize);
			if (inAddHere <= nItems)
				BlockMoveData( sourceP, va_NthItemPtr(inAddHere + inCount,data,itemSize),
								(nItems - inAddHere + 1) * itemSize );
			// Optionally zero the added items
			if (inZero) VAZeroMemory( sourceP, inCount * itemSize );
			nItems += inCount;
			va_N(inArray) = nItems;
		}
	}
	return sourceP;
}

//===============================
// Returns a pointer to the first element added.
// In case of error, returns nil.
VA_PROC void* VAAddToEnd( UInt32 inCount, Boolean inZero, VARef inArray )
{
	return VAAdd( inCount, va_N(inArray) + 1, inZero, inArray );
}

//===============================
VA_PROC OSErr VAInsert( UInt32 inAddHere, void* inData, VARef inArray )
{
	OSErr	err = noErr;
	
	if (inAddHere > va_N(inArray) || !inAddHere) err = inputOutOfBounds;
	else
	{
		void*	p;
		
		p = VAAdd( 1, inAddHere, false, inArray );
		if (!p) err = memFullErr;
		else (void)VASetNthItem( inAddHere, inData, inArray );
	}
	return err;
}

//===============================
VA_PROC OSErr VAInsertAtEnd( void* inData, VARef inArray )
{
	OSErr	err = noErr;
	void*	p;
	
	p = VAAddToEnd( 1, false, inArray );
	if (!p) err = memFullErr;
	else err = VASetNthItem( va_N(inArray), inData, inArray );
	return err;
}

//===============================	
VA_PROC void VARemove( UInt32 inCount, UInt32 inItem, VARef inArray )
{
	UInt32		nItems;
	
	nItems = va_N(inArray);
	_AssertStr( inItem != 0, "\pError: cannot delete element #0." );
	_AssertStr( inCount != 0, "\pNotice: deleting 0 items has no effect." );
	_AssertStr( inCount <= nItems, "\pNotice: trying to delete more items than exist." );
	_AssertStr( inItem + inCount - 1 <= nItems, "\pNotice: trying to delete more items than exist." );
	if (inCount && nItems && inItem)
	{
		Ptr			data;
		UInt32		entriesBelow, entriesAbove, itemSize, neededSize, newNItems;
		
		if (inCount > nItems) inCount = nItems;
		if (inItem + inCount - 1 > nItems) inCount = nItems - inItem + 1;
		itemSize = va_S(inArray);
		data = va_D(inArray);
		entriesBelow = inItem - 1;
		entriesAbove = nItems - inCount - entriesBelow;
		// Move to target of shift
		data += entriesBelow * itemSize;
		BlockMoveData(	data + (inCount * itemSize), data, entriesAbove * itemSize );
		newNItems = nItems - inCount;
		neededSize = newNItems * itemSize + sizeof(VARec) - 1;
		SetHandleSize( (Handle)inArray, neededSize );
		va_N(inArray) = newNItems;
	}
}

//===============================	
VA_PROC void VARemoveAll( VARef inArray )
{
	SetHandleSize( (Handle)inArray, sizeof(VARec) );
	va_N(inArray) = 0;
}

//===============================
// This is the PowerPlant-style move. It is as though the entries are deleted and
//	reinserted at inDest position, provided it could be done without resizing the array.
OSErr VAMove( UInt32 inCount, UInt32 inSource, UInt32 inDest, VARef inArray )
{
	OSErr		err = noErr;
	
	if (inSource != inDest && inCount)
	{
		UInt32	moveAfter;
		
		if (inDest < inSource) moveAfter = inDest - 1;
		else moveAfter = inCount + inDest - 1;
		err = VAMoveAfter( inCount, inSource, moveAfter, inArray );
	}
	return err;
}


#define vaCheaterStorageBytes	256
//===============================
// This is where a specific entry is the target of movement.
// We try to copy inCount entries, starting with inSource, so that inSource becomes
//	the entry after inDest.
// It's fine to pass 0 for inDest. The items get moved to the front of the array.
VA_PROC OSErr VAMoveAfter( UInt32 inCount, UInt32 inSource, UInt32 inDest, VARef inArray )
{
	UInt32		nItems;
	OSErr		err;
	
	nItems = va_N(inArray);
	err = VAValidateMoveParameters( inCount, inSource, inDest, nItems );
	if (!err)
	{
		UInt32		itemSize, source, dest, count, left, right;
		char		cheaterStorage[vaCheaterStorageBytes];
		
		itemSize = va_S(inArray);
		// Set things up so we are shifting source upwards in memory.
		if (inSource > inDest)
		{
			source = inDest + 1;
			dest = inSource + inCount - 1;
			count = inSource - inDest - 1;
		}
		else
		{
			source = inSource;
			dest = inDest;
			count = inCount;
		}
		// left is the count of (ordered) entries waiting to be moved up into position.
		// right count also refers to ordered entries; they will be swapped down.
		left = count;
		right = dest + 1 - source - count;
		do
		{	
			// Cheat against possible degenerate circumstances such as left = 1, right = 4000.
			if (left * itemSize <= vaCheaterStorageBytes)
			{
				// Move left into storage, move right down by nEntries, move left into place.
				BlockMoveData( VANthItemPtr( source, inArray ), &cheaterStorage[0], left * itemSize );
				BlockMoveData( VANthItemPtr( source + left, inArray ), VANthItemPtr( source, inArray ), right * itemSize );
				BlockMoveData( &cheaterStorage[0], VANthItemPtr( source + right, inArray ), left * itemSize );
				left = 0; // We're done!
			}
			else
			{
				UInt32	s1, s2; // The indices of the first entries to be swapped.
								// The swapping starts: swap s1 and s2, then s1 - 1 and s2 - 1,
								//	till the number of swaps for this iteration is done.
				UInt32	bytes;
				
				if (left == right)
				{
					s1 = source;
					s2 = source + left;
					bytes = left * itemSize;
					left = 0; // We're done!
				}
				else if (left < right)
				{
					s1 = source;
					s2 = source + right; // This looks wrong but it ain't!
					bytes = left * itemSize;
					right -= left;
				}
				else
				{
					s1 = source + left - right;
					s2 = source + left;
					bytes = right * itemSize;
					left -= right;
				}
				VASwapMemory( VANthItemPtr( s1, inArray ), VANthItemPtr( s2, inArray ), bytes );
			}
		} while (left);
	}
	return err;
}


//===============================
// Validates the input to VAMoveAfter.
// inNArrayItems is VACount( array ).
VA_PROC OSErr VAValidateMoveParameters( UInt32 inCount, UInt32 inSource, UInt32 inDest,
										UInt32 inNArrayItems )
{
	OSErr		err = noErr;
	
	// Moving to after inSource or inSource - 1, or moving 0 entries, results in no movement.
	if (inSource == inDest || inSource - 1 == inDest || !inCount) err = vaNoMovementNecessaryErr;
	else
	{
		_AssertStr( inSource && inSource <= inNArrayItems, "\pError: source does not exist." );
		if (!inSource || inSource > inNArrayItems) err = inputOutOfBounds;
		_AssertStr( inDest <= inNArrayItems, "\pError: destination does not exist." );
		if (inDest > inNArrayItems) err = inputOutOfBounds;
		// Make sure count is not greater than or equal to the total number of items.
		_AssertStr( inCount < inNArrayItems, "\pError: trying to move too many items." );
		if (inCount >= inNArrayItems) err = inputOutOfBounds;
		if (!err)
		{
			UInt32	maxCount;
			
			// Check the inCount parameter to make sure we're not trying to move too many items.
			// The maximum count is the moved entry and all above it.
			maxCount = inNArrayItems - inSource + 1;
			_AssertStr( inCount <= maxCount, "\pError: trying to move too many items." );
			if (inCount > maxCount) err = inputOutOfBounds;
			else
			{
				// Also, if the destination is greater than the source, we can't move the
				//	destination along with the source items.
				// If we have {abcdefgh} and we move 3 entries starting with #2 (b),
				//	the first entry we can move after is #5 (e). first = source + count.
				if (inDest > inSource)
				{
					_AssertStr( inDest >= inSource + inCount, "\pError: trying to move items into itself." );
					if (inDest < inSource + inCount) err = inputOutOfBounds;
				}
			}
		}
	}
	return err;
}

//===============================
VA_PROC OSErr VASwap( UInt32 inItem1, UInt32 inItem2, VARef inArray )
{
	OSErr	err = noErr;
	UInt32	nItems;
	
	nItems = va_N(inArray);
	if (!inItem1 || inItem1 > nItems) err = inputOutOfBounds;
	else
	{
		if (!inItem2 || inItem2 > nItems) err = inputOutOfBounds;
		else if (inItem1 != inItem2)
		{
			Ptr			d, p1, p2;
			UInt32		size;
			
			d = va_D(inArray);
			size = va_S(inArray);
			p1 = va_NthItemPtr(inItem1,d,size);
			p2 = va_NthItemPtr(inItem2,d,size);
			VASwapMemory( p1, p2, size );
		}
	}
	return err;
}

//===============================
// If your search proc moves memory, you have to lock the array!!
// Requires a sorted array!
// Returns nthItem; zero if no match.
// If no match and 'next' pointer is valid, returns the next element or zero if search item
//	would be the last one.
VA_PROC UInt32 VASearch( VASearchProcPtr inProc, UInt32 inRefCon, void* inSearchData,
						UInt32* outNext, VARef inArray )
{
	UInt32		matchedItem = 0, next = 0;
	
	if (inProc)
	{
		UInt32		nItems = va_N(inArray);
		
		if (nItems)
		{
			UInt32		index, itemSize;
			SInt32		left = 0, right, response;
			Ptr			data;
			
			itemSize = va_S(inArray);
			right = nItems - 1;
			data = va_D(inArray);
			// left starts at 0 and right is at least 0, so it's safe to use a do while loop.
			do
			{
				index = (left + right) / 2;
				response = (*inProc)( va_NthItemPtr(index+1,data,itemSize), inSearchData, inRefCon );
				if (response == vaEqual)
				{
					matchedItem = index + 1;
					// Return the next one unless the matched item is the last item
					if (matchedItem != nItems) next = matchedItem + 1;
					break;
				}
				// Search item is greater
				if (response == vaSearchDataIsGreater) left = index + 1;
				else right = index - 1;
			} while (right >= left);
			if (!matchedItem)
			{
				// If the response was that the search data was greater for the last index,
				//	get the one after the index if it exists.
				if (response == vaSearchDataIsGreater)
				{
					if (index + 2 > nItems) next = 0;
					else next = index + 2;
				}
				// Otherwise this is the one we want
				else next = index + 1;
			}
		} // 1 or more items in array.
	} // Comparison proc not nil.
	if (outNext) *outNext = next;
	return matchedItem;
}


//===============================
VA_PROC UInt32 VABruteForceSearch( VABruteForceSearchProcPtr inProc, UInt32 inRefCon,
									void* inSearchData, VARef inArray )
{
	UInt32		matchedItem = 0;
	
	if (inProc)
	{
		UInt32		index, nItems, itemSize;
		Ptr			data;
		
		nItems = va_N(inArray);
		itemSize = va_S(inArray);
		data = va_D(inArray);
		for (index = 1; index <= nItems; index++)
		{
			SInt32	response;
			
			response = (*inProc)( va_NthItemPtr(index,data,itemSize), inSearchData, inRefCon );
			if (response == vaEqual)
			{
				matchedItem = index;
				break;
			}
		}
	}
	return matchedItem;
}

//===============================
// Heapsort the array using the passed-in comparison routine.
// The inRefCon parameter has nothing to do with the array refCon; it's passed to
//	the comparison callback in case extra data is needed.
VA_PROC void VASort( VASortProcPtr inProc, UInt32 inRefCon, VARef inArray )
{
	UInt32	nItems, itemSize;
	Ptr		data;
	UInt8	state;
	
	state = HGetState( (Handle)inArray );
	HLock( (Handle)inArray );
	data = va_D(inArray);
	nItems = va_N(inArray);
	itemSize = va_S(inArray);
    switch (itemSize)
    {
		case sizeof(UInt16):
		local_VAHeapsort16( (UInt16*)data, nItems, inProc, inRefCon );
		break;

		case sizeof(UInt32):
		local_VAHeapsort32( (UInt32*)data, nItems, inProc, inRefCon );
		break;

		default:
		local_VAHeapsortOther( data, nItems, itemSize, inProc, inRefCon );
		break;
    }
    HSetState( (Handle)inArray, state );
}

//===============================
static void local_VAReheap16( UInt16* arrayP, UInt32 left, UInt32 right,
								VASortProcPtr proc, UInt32 refCon )
{
	UInt32	j, j2;
	
	j = left;
	j2 = j << 1;
	while (j2 < right)
	{
		j = j2;
		if ((*proc)((Ptr)(va_NthItemPtr16(j+1,arrayP)), (Ptr)(va_NthItemPtr16(j,arrayP)),
					refCon) >= vaFirstIsGreater ) ++j;
		j2 = j << 1;
	}
	if (j2 == right) j = right;
	while ((left < j) && (*proc)((Ptr)(va_NthItemPtr16(left,arrayP)),
									(Ptr)(va_NthItemPtr16(j,arrayP)),
									refCon) >= vaFirstIsGreater ) j >>= 1;
	if (left < j)
	{
	    do
	    {
			UInt16	temp;
			
			temp = *va_NthItemPtr16(left,arrayP);
			*va_NthItemPtr16(left,arrayP) = *va_NthItemPtr16(j,arrayP);
			*va_NthItemPtr16(j,arrayP) = temp;
	        j >>= 1;
	    }
	    while ( left < j );
	}
}

//===============================
static void local_VAReheap32( UInt32* arrayP, UInt32 left, UInt32 right,
								VASortProcPtr proc, UInt32 refCon )
{
	UInt32	j, j2;
	
	j = left;
	j2 = j << 1;
	while (j2 < right)
	{
		j = j2;
		if ((*proc)((Ptr)(va_NthItemPtr32(j+1,arrayP)), (Ptr)(va_NthItemPtr32(j,arrayP)),
							refCon) >= vaFirstIsGreater ) ++j;
		j2 = j << 1;
	}
	if (j2 == right) j = right;
	while ((left < j) && (*proc)((Ptr)(va_NthItemPtr32(left,arrayP)),
									(Ptr)(va_NthItemPtr32(j,arrayP)),
									refCon) >= vaFirstIsGreater ) j >>= 1;
	if (left < j)
	{
	    do
	    {
		    UInt32	temp;

		    temp = *va_NthItemPtr32(left,arrayP);
			*va_NthItemPtr32(left,arrayP) = *va_NthItemPtr32(j,arrayP);
			*va_NthItemPtr32(j,arrayP) = temp;
	        j >>= 1;
	    }
	    while ( left < j );
	}
}

//===============================
static void local_VAReheapOther( Ptr arrayP, UInt32 itemSize, UInt32 left, UInt32 right,
								VASortProcPtr proc, UInt32 refCon )
{
    UInt32	j, j2;
	
	j = left;
	j2 = j << 1;
	while (j2 < right)
	{
		j = j2;
		if ((*proc)(va_NthItemPtr(j+1,arrayP,itemSize), va_NthItemPtr(j,arrayP,itemSize),
					refCon) >= vaFirstIsGreater ) ++j;
		j2 = j << 1;
	}
	if (j2 == right) j = right;

	while ((left < j) && (*proc)(va_NthItemPtr(left,arrayP,itemSize),
									va_NthItemPtr(j,arrayP,itemSize),
									refCon) >= vaFirstIsGreater ) j >>= 1;

	if (left < j)
	{
		Ptr	leftP = va_NthItemPtr(left,arrayP,itemSize);
	    do
	    {
			Ptr	rightP = va_NthItemPtr(j,arrayP,itemSize);

			VASwapMemory( leftP, rightP, itemSize );
	        j >>= 1;
	    }
	    while ( left < j );
	}
}

//===============================
static void local_VAHeapsort16( UInt16* arrayP, UInt32 nItems, VASortProcPtr proc,
								UInt32 refCon )
{
	UInt32	i;

	for (i = nItems >> 1; i >= 1; --i) local_VAReheap16( arrayP, i, nItems, proc, refCon );

	for ( i = nItems; i >= 2; --i )
	{
		UInt16	temp;
		
		temp = *va_NthItemPtr16(1,arrayP);
		*va_NthItemPtr16(1,arrayP) = *va_NthItemPtr16(i,arrayP);
		*va_NthItemPtr16(i,arrayP) = temp;
		local_VAReheap16( arrayP, 1, i - 1, proc, refCon );
	}
}

//===============================
static void local_VAHeapsort32( UInt32* arrayP, UInt32 nItems, VASortProcPtr proc,
								UInt32 refCon)
{
	UInt32	i;

	for (i = nItems >> 1; i >= 1; --i) local_VAReheap32( arrayP, i, nItems, proc, refCon );
	for ( i = nItems; i >= 2; --i )
	{
		UInt32	temp;

		temp = *va_NthItemPtr32(1,arrayP);
		*va_NthItemPtr32(1,arrayP) = *va_NthItemPtr32(i,arrayP);
		*va_NthItemPtr32(i,arrayP) = temp;
		local_VAReheap32( arrayP, 1, i - 1, proc, refCon );
	}
}

//===============================
static void local_VAHeapsortOther( Ptr arrayP, UInt32 nItems, UInt32 itemSize,
									VASortProcPtr proc, UInt32 refCon )
{
	UInt32	i;
	Ptr		leftP;
	
	leftP = va_NthItemPtr(1,arrayP,itemSize);
	for (i = nItems >> 1; i >= 1; --i)
		local_VAReheapOther( arrayP, itemSize, i, nItems, proc, refCon );
	for (i = nItems; i >= 2; --i)
	{
		Ptr		rightP;
		
		rightP = va_NthItemPtr(i,arrayP,itemSize);
		VASwapMemory( leftP, rightP, itemSize );
		local_VAReheapOther( arrayP, itemSize, 1, i - 1, proc, refCon );
	}
}


#if TARGET_CPU_PPC
#if !defined __MRC__

	static const double gZeroDouble = 0.0;
	#define _VAPPCZeroMemory VAZeroMemory

//===============================
// p is in r3, bytes in r4, zero in r0.
asm void _VAPPCZeroMemory( register Ptr inPtr, register UInt32 inBytes )
{
	#pragma unused (inPtr,inBytes)
	cmplwi		r3,0			// Check for nil pointer
	beqlr						// Return if equal (to nil)
	li			r0,0			// Load zero value
	cmplwi		r4,16			// If less than 16 bytes to copy...
	ble			@SetupCopyByte	//	copy byte by bye at end
@AlignByte
	rlwinm.		r6,r3,0,31,31	// Check bit 31 (ones place)
	beq			@AlignShort		// If bit 31 is clear, try to align to short
	stb			r0,0(r3)		// Store zero byte to ptr
	addi		r3,r3,1			// Increment ptr
	subi		r4,r4,1			// Decrement bytes to copy
@AlignShort
	rlwinm.		r6,r3,0,30,30	// Check bit 30 (twos place)
	beq			@AlignLong		// If bit 30 is clear, try to align to long
	sth			r0,0(r3)		// Store zero short to ptr
	addi		r3,r3,2			// Increment ptr
	subi		r4,r4,2			// Decrement bytes to copy
@AlignLong
	rlwinm.		r6,r3,0,29,29	// Check bit 29 (fours place)
	beq			@SetupCopyDouble// If bit 29 is clear, try to align to double
	stw			r0,0(r3)		// Store zero long to ptr
	addi		r3,r3,4			// Increment ptr
	subi		r4,r4,4			// Decrement bytes to copy
@SetupCopyDouble
	lfd			fp0,gZeroDouble	// Set fp register 0 to zero value. There is no
								//	floating-point load immediate instruction (too bad) so we
								//	have to use a global.
	subi		r3,r3,8			// Back up so store double with update instruction (stfdu)
								//	can add 8 to ptr to get store location
@CheckDouble
	cmplwi		r4,8			// At least 8 bytes left to copy?
	blt			@DoneWithDouble	// Clean up if not
@ZeroDouble
	stfdu		fp0,8(r3)		// Store double to memory offset by immediate value
								//	(in this case, 8) and increment r3 by that amount
	subi		r4,r4,8			// Decrement bytes to copy
	b			@CheckDouble	// Loop over double-storing code
@DoneWithDouble
	addi		r3,r3,8			// Reset r3 to point to next byte to be zeroed
@SetupCopyByte
	subi		r3,r3,1			// Do the same trick as with doubles, except we
								//	only 'autoincrement' by 1
@CheckByte
	cmplwi		r4,0			// Zero bytes left to copy?
	beqlr						// Return if so
@CopyByte
	stbu		r0,1(r3)		// Store byte to memory offset by the immediate value (1)
								//	and autoincrement by 1. This is like saying *(++p) = 0;
	subi		r4,r4,1			// Decrement bytes left
	b			@CheckByte		// Loop over byte-storing (exit condition is there)
}

#endif // Not __MRC__
#else // Not PowerPC

	#define _VA68kZeroMemory VAZeroMemory

//===============================
// 68k alignment rule: load/store 2 or 4 bytes has to be at even address. That's all.
void _VA68kZeroMemory( Ptr inPtr, UInt32 inBytes )
{
	_AssertStr(inPtr != nil,"\pError: nil pointer.");
	if (inPtr)
	{
		_AssertStr(inBytes > 0, "\pNotice: tried to clear zero bytes.");
		if (inBytes >= 5)
		{
			if ((UInt32)inPtr & 0x00000001)
			{
				*inPtr = 0;
				inPtr++;
				inBytes--;
			} // P is now guaranteed to be even
			while (inBytes >= 4)
			{
				*(UInt32*)inPtr = 0L;
				inPtr += 4;
				inBytes -= 4;
			}
		}
		// Take care of the rest
		while (inBytes > 0)
		{
			*inPtr = 0;
			inPtr++;
			inBytes--;
		}
	}
}

#endif // 68K or PowerPC

#if TARGET_CPU_PPC
#if !defined __MRC__
	
	#define _VAPPCSwapMemory VASwapMemory

//===============================
// r0:			rlwinm result, storage
// r3:			inPtr1
// r4:			inPtr2
// r5:			inBytes
// r6:			storage
// fp1, fp2:	storage
asm void _VAPPCSwapMemory( register Ptr inPtr1, register Ptr inPtr2, register UInt32 inBytes )
{
	#pragma unused (inPtr1,inPtr2, inBytes)
	cmplw		r3,r4			// See if blocks are the same
	beqlr						// Return if equal
	cmplwi		r4,16			// If less than 16 bytes to copy...
	blt			@Bytes			//	copy byte by bye at end
	rlwinm.		r0,r3,0,31,31	// Check bit 31 (ones place)
	beq			@CheckPtr2HalfWordAligned
								// If bit 31 is set, try to align to short
	lbz			r0,0(r3)		// Get a byte from inPtr1
	lbz			r6,0(r4)		// Get a byte from inPtr2
	stb			r0,0(r4)		// Store byte in inPtr2
	stb			r6,0(r3)		// Store byte in inPtr1
	addi		r3,r3,1			// Increment pointers...
	addi		r4,r4,1
	subi		r5,r5,1			// Decrement bytes to copy
@CheckPtr2HalfWordAligned
	rlwinm.		r0,r4,0,31,31	// Check bit 31 (ones place) for inPtr2.
	bne			@Bytes			// If odd, we can't halfword-align the two pointers
	rlwinm.		r0,r3,0,30,30	// Check bit 30 (twos place)
	beq			@CheckPtr2WordAligned
								// If bit 30 is set, try to align to word
	lhz			r0,0(r3)		// Get a half word from inPtr1
	lhz			r6,0(r4)		// Get a half word from inPtr2
	sth			r0,0(r4)		// Store half word in inPtr2
	sth			r6,0(r3)		// Store half word in inPtr1
	addi		r3,r3,2			// Increment pointers...
	addi		r4,r4,2
	subi		r5,r5,2			// Decrement bytes to copy
@CheckPtr2WordAligned
	rlwinm.		r0,r4,0,30,30	// Check bit 30 (twos place) for inPtr2.
	bne			@HalfWords		// If not multiple of 4, we can't word-align the two pointers
	rlwinm.		r0,r3,0,29,29	// Check bit 29 (fours place)
	beq			@CheckPtr2DoubleAligned
								// If bit 29 is set, try to align to double
	lwz			r0,0(r3)		// Get a word from inPtr1
	lwz			r6,0(r4)		// Get a word from inPtr2
	stw			r0,0(r4)		// Store word in inPtr2
	stw			r6,0(r3)		// Store word in inPtr1
	addi		r3,r3,4			// Increment pointers...
	addi		r4,r4,4
	subi		r5,r5,4			// Decrement bytes to copy
@CheckPtr2DoubleAligned
	rlwinm.		r0,r4,0,29,29	// Check bit 29 (fours place) for inPtr2.
	bne			@Words			// If not multiple of 8, we can't double-align the two pointers
@Doubles
	cmplwi		r5,8			// Compare remaining bytes to 8
	blt			@Words			// If less than 8 bytes, we can't copy a double
	lfd			fp0,0(r3)		// Get a double from inPtr1
	lfd			fp1,0(r4)		// Get a double from inPtr2
	stfd		fp0,0(r4)		// Store double in inPtr2
	stfd		fp1,0(r3)		// Store double in inPtr1
	addi		r3,r3,8			// Increment pointers...
	addi		r4,r4,8
	subi		r5,r5,8			// Decrement bytes to copy
	b			@Doubles		// Go back to check bytes left
@Words
	cmplwi		r5,4			// Compare remaining bytes to 4
	blt			@HalfWords		// If less than 4 bytes, we can't copy a word
	lwz			r0,0(r3)		// Get a word from inPtr1
	lwz			r6,0(r4)		// Get a word from inPtr2
	stw			r0,0(r4)		// Store word in inPtr2
	stw			r6,0(r3)		// Store word in inPtr1
	addi		r3,r3,4			// Increment pointers...
	addi		r4,r4,4
	subi		r5,r5,4			// Decrement bytes to copy
	b			@Words			// Go back to check bytes left
@HalfWords
	cmplwi		r5,2			// Compare remaining bytes to 2
	blt			@Bytes			// If less than 2 bytes, we can't copy a half word
	lhz			r0,0(r3)		// Get a half word from inPtr1
	lhz			r6,0(r4)		// Get a half word from inPtr2
	sth			r0,0(r4)		// Store half word in inPtr2
	sth			r6,0(r3)		// Store half word in inPtr1
	addi		r3,r3,2			// Increment pointers...
	addi		r4,r4,2
	subi		r5,r5,2			// Decrement bytes to copy
	b			@HalfWords		// Go back to check bytes left
@Bytes
	subi		r3,r3,1			// Set up for the load/store with update instructions
	subi		r4,r4,1			//	we only do this once
@BytesNowSetup
	cmplwi		r5,1			// Compare remaining bytes to 1
	bltlr						// EXIT if no bytes left
	lbzu		r0,0(r3)		// Increment inPtr1 and get a byte
	lbzu		r6,0(r4)		// Increment inPtr2 and get a byte
	stb			r0,0(r4)		// Store byte in inPtr2
	stb			r6,0(r3)		// Store byte in inPtr1
	subi		r5,r5,1			// Decrement bytes to copy
	b			@BytesNowSetup	// Go back to check bytes left
}

#endif // Not __MRC__
#else // Not PowerPC

	#define _VA68kSwapMemory VASwapMemory
	
	#define va_SwapByte		{	byteStorage = *inPtr1; *inPtr1 = *inPtr2; *inPtr2 = byteStorage; \
								inPtr1++; inPtr2++; inBytes--; }
	
	#define va_SwapHalfWord	{	halfWordStorage = *(UInt16*)inPtr1; \
								halfWordPtr = (UInt16*)inPtr1; *halfWordPtr = *(UInt16*)inPtr2; \
								halfWordPtr = (UInt16*)inPtr2; *halfWordPtr = halfWordStorage; \
								inPtr1 += sizeof(UInt16); inPtr2 += sizeof(UInt16); \
								inBytes -= sizeof(UInt16); }
	
	#define va_SwapWord		{	wordStorage = *(UInt32*)inPtr1; \
								wordPtr = (UInt32*)inPtr1; *wordPtr = *(UInt32*)inPtr2; \
								wordPtr = (UInt32*)inPtr2; *wordPtr = wordStorage; \
								inPtr1 += sizeof(UInt32); inPtr2 += sizeof(UInt32); \
								inBytes -= sizeof(UInt32); }

//===============================
// Does not check for overlap!
void _VA68kSwapMemory( Ptr inPtr1, Ptr inPtr2, UInt32 inBytes )
{
	UInt8	byteStorage;
	
	_AssertStr(inPtr1 != nil,"\pError: nil pointer.");
	_AssertStr(inPtr2 != nil,"\pError: nil pointer.");
	if (inPtr1 && inPtr2)
	{
		_AssertStr(inBytes > 0, "\pNotice: tried to swap zero bytes.");
		_AssertStr(inPtr1 != inPtr2, "\pNotice: tried to swap identical blocks.");
		if (inPtr1 != inPtr2)
		{
			if (inBytes >= 8)
			{
				if ((UInt32)inPtr1 & 1) va_SwapByte;
				if (!(UInt32)inPtr2 & 1)
				{
					UInt16	halfWordStorage, *halfWordPtr;
					
					if ((UInt32)inPtr1 & 2) va_SwapHalfWord;
					if (!(UInt32)inPtr2 & 2)
					{
						UInt32	wordStorage, *wordPtr;
						
						while (inBytes <= sizeof(UInt32)) va_SwapWord;
					}
					while (inBytes <= sizeof(UInt16)) va_SwapHalfWord;
				}
			}
			while (inBytes) va_SwapByte;
		}
	}
}

#endif // 68K or PowerPC

//===============================
VA_PROC void* VANthItemPtr( UInt32 inItem, VARef inArray )
{
	UInt32	nItems;
	void*	p;
	
	nItems = va_N(inArray);
	if (inItem > nItems || !inItem) p = nil;
	else p = va_NthItemPtr(inItem,va_D(inArray),va_S(inArray));
	return p;
}

//===============================
VA_PROC void* VALastItemPtr( VARef inArray )
{
	UInt32	nItems;
	void*	p;
	
	nItems = va_N(inArray);
	if (!nItems) p = nil;
	else p = va_NthItemPtr(nItems,va_D(inArray),va_S(inArray));
	return p;
}


//===============================
VA_PROC OSErr VAGetNthItem( UInt32 inItem, void* outData, VARef inArray )
{
	UInt32	nItems;
	OSErr	err;
	
	nItems = va_N(inArray);
	if (inItem <= nItems && inItem)
	{
		UInt32	itemSize;
		Ptr		itemPtr;
		
		itemSize = va_S(inArray);
		itemPtr = va_NthItemPtr(inItem,va_D(inArray),itemSize);
		BlockMoveData( itemPtr, outData, itemSize );
		err = noErr;
	}
	else err = inputOutOfBounds;
	return err;
}

//===============================
VA_PROC OSErr VASetNthItem( UInt32 inItem, void* inData, VARef inArray )
{
	UInt32	nItems;
	OSErr	err;
	
	nItems = va_N(inArray);
	if (inItem <= nItems && inItem)
	{
		UInt32	itemSize;
		Ptr		itemPtr;
		
		itemSize = va_S(inArray);
		itemPtr = va_NthItemPtr(inItem,va_D(inArray),itemSize);
		BlockMoveData( inData, itemPtr, itemSize );
		err = noErr;
	}
	else err = inputOutOfBounds;
	return err;
}


//===============================
// If inTargetArray is nil, creates a new one. Otherwise makes inTargetArray identical.
// Preserves ownerCount. Sets it to 1 if creating a new array.
VA_PROC VARef VACopy( VARef inSourceArray, VARef inTargetArray )
{
	VARef		newArray;
	UInt32		ownerCount, size;
	OSErr		err;
	
	size = GetHandleSize( (Handle)inSourceArray );
	if (inTargetArray)
	{
		ownerCount = va_OC(inTargetArray);
		newArray = inTargetArray;
		SetHandleSize( (Handle)inTargetArray, size );
		err = MemError();
	}
	else
	{
		ownerCount = 1;
		newArray = (VARef)NewHandle( size );
		if (!newArray) err = memFullErr;
		else err = noErr;
	}
	
	if (!err)
	{
		BlockMoveData( *inSourceArray, *newArray, size );
		va_OC(newArray) = ownerCount;
	}
	return newArray;
}

//===============================
// Copies the first n entries of the array.
// Should expand this to allow copying n entries from anywhere in the array:
// VARef VACopyPart( UInt32 inCount, UInt32 inFirstEntry, VARef inArray )
VA_PROC VARef VACopyPart( UInt32 inCount, VARef inArray )
{
	VARef		newArray;
	UInt32		nItems, itemSize, refCon;
	
	nItems = va_N(inArray);
	itemSize = va_S(inArray);
	refCon = va_RC(inArray);
	if (inCount > nItems) inCount = nItems;
	newArray = VANew( inCount, itemSize, false, refCon );
	if (newArray) BlockMoveData( va_D(inArray), va_D(newArray), inCount * itemSize );
	return newArray;
}

//===============================
VA_PROC VARef VAClone( VARef inArray )
{
	va_OC(inArray)++;
	return inArray;
}

//===============================
VA_PROC OSErr VAConcatenate( VARef inSource, VARef inDest )
{
	UInt32	sourceSize, destSize;
	OSErr	err;
	
	sourceSize = va_S(inSource);
	destSize = va_S(inDest);
	// Make sure data items are the same size.
	if (sourceSize == destSize)
	{
		UInt32	sourceCount, destCount;
		Ptr		source;
		
		sourceCount = va_N(inSource);
		destCount = va_N(inDest);
		source = va_NthItemPtr(1,va_D(inSource),sourceSize);
		err = PtrAndHand( source, (Handle)inDest, sourceSize * sourceCount );
		if (!err) va_N(inDest) += sourceCount;
	}
	else err = paramErr;
	return err;
}

//===============================
VA_PROC void VAIterate( VAIteratorPtr inProc, void* ioUserData, VARef inArray )
{
	Ptr		data;
	UInt32	itemSize, nItems, i;
	SInt8	state;
	
	state = HGetState( (Handle)inArray );
	HLock( (Handle)inArray );
	itemSize = va_S(inArray);
	nItems = va_N(inArray);
	data = va_NthItemPtr(1,va_D(inArray),itemSize);
	for (i = 1; i <= nItems; i++, data += itemSize)
	{
		VAIteratorResult result;
		
		result = (*inProc)( data, ioUserData );
		if (result == vaIteratorStop) break;
	}
	HSetState( (Handle)inArray, state );
}

#pragma mark -
//===============================
VA_PROC void VAInternalSort( Size inSize, Boolean inSigned, UInt32 inOffset, VARef inArray )
{
	VASortProcPtr	proc;
	
	switch (inSize)
	{
		case 1:
		proc = (inSigned) ? &local_VASignedByteSortProc : &local_VAByteSortProc;
		break;
		
		case 2:
		proc = (inSigned) ? &local_VASignedHalfWordSortProc : &local_VAHalfWordSortProc;
		break;
		
		case 4:
		proc = (inSigned) ? &local_VASignedWordSortProc : &local_VAWordSortProc;
		break;
		
		default:
		{
			VADataSortParam	param;
			
			param.size = inSize;
			param.offset = inOffset;
			inOffset = (UInt32)&param;
			proc = (VASortProcPtr)&local_VADataSortProc;
		}
		break;
	}
	VASort( proc, inOffset, inArray );
}

//===============================
static VAComparisonResult local_VAByteSortProc( const Ptr data1P, const Ptr data2P, UInt32 inOffset )
{
	UInt8	*p1, *p2;
	
	p1 = (UInt8*)(data1P + inOffset);
	p2 = (UInt8*)(data2P + inOffset);
	if (*p1 > *p2) return vaFirstIsGreater;
	if (*p1 < *p2) return vaFirstIsLess;
	return vaEqual;
}

//===============================
static VAComparisonResult local_VASignedByteSortProc( const Ptr data1P, const Ptr data2P, UInt32 inOffset )
{
	SInt8	*p1, *p2;
	
	p1 = (SInt8*)(data1P + inOffset);
	p2 = (SInt8*)(data2P + inOffset);
	if (*p1 > *p2) return vaFirstIsGreater;
	if (*p1 < *p2) return vaFirstIsLess;
	return vaEqual;
}

//===============================
static VAComparisonResult local_VAHalfWordSortProc( const Ptr data1P, const Ptr data2P, UInt32 inOffset )
{
	UInt16	*p1, *p2;
	
	p1 = (UInt16*)(data1P + inOffset);
	p2 = (UInt16*)(data2P + inOffset);
	if (*p1 > *p2) return vaFirstIsGreater;
	if (*p1 < *p2) return vaFirstIsLess;
	return vaEqual;
}

//===============================
static VAComparisonResult local_VASignedHalfWordSortProc( const Ptr data1P, const Ptr data2P, UInt32 inOffset )
{
	SInt16	*p1, *p2;
	
	p1 = (SInt16*)(data1P + inOffset);
	p2 = (SInt16*)(data2P + inOffset);
	if (*p1 > *p2) return vaFirstIsGreater;
	if (*p1 < *p2) return vaFirstIsLess;
	return vaEqual;
}

//===============================
static VAComparisonResult local_VAWordSortProc( const Ptr data1P, const Ptr data2P, UInt32 inOffset )
{
	UInt32	*p1, *p2;
	
	p1 = (UInt32*)(data1P + inOffset);
	p2 = (UInt32*)(data2P + inOffset);
	if (*p1 > *p2) return vaFirstIsGreater;
	if (*p1 < *p2) return vaFirstIsLess;
	return vaEqual;
}

//===============================
static VAComparisonResult local_VASignedWordSortProc( const Ptr data1P, const Ptr data2P, UInt32 inOffset )
{
	SInt32	*p1, *p2;
	
	p1 = (SInt32*)(data1P + inOffset);
	p2 = (SInt32*)(data2P + inOffset);
	if (*p1 > *p2) return vaFirstIsGreater;
	if (*p1 < *p2) return vaFirstIsLess;
	return vaEqual;
}

//===============================
static VAComparisonResult local_VADataSortProc( const Ptr data1P, const Ptr data2P,
												VADataSortParam* inSortData )
{
	Size	i;
	UInt8	*p1, *p2;
	
	p1 = (UInt8*)(data1P + inSortData->offset);
	p2 = (UInt8*)(data2P + inSortData->offset);
	for (i = 1; i <= inSortData->size; i++, p1++, p2++)
	{
		if (*p1 > *p2) return vaFirstIsGreater;
		if (*p1 < *p2) return vaFirstIsLess;
	}
	return vaEqual;
}

//===============================
VA_PROC UInt32 VAInternalSearch( Size inSize, Boolean inSigned, UInt32 inOffset,
								void* inSearchData, UInt32* outNext, VARef inArray )
{
	UInt32			nthItem = 0;
	VASearchProcPtr	proc;
	
	if (outNext) *outNext = 0;
	switch (inSize)
	{
		case 1:
		proc = (inSigned) ? (VASearchProcPtr)&local_VASignedByteSearchProc :
							(VASearchProcPtr)&local_VAByteSearchProc;
		break;
		
		case 2:
		proc = (inSigned) ? (VASearchProcPtr)&local_VASignedHalfWordSearchProc :
							(VASearchProcPtr)&local_VAHalfWordSearchProc;
		break;
		
		case 4:
		proc = (inSigned) ? (VASearchProcPtr)&local_VASignedWordSearchProc :
							(VASearchProcPtr)&local_VAWordSearchProc;
		break;
		
		default:
		{
			VADataSortParam	param;
			
			param.size = inSize;
			param.offset = inOffset;
			inOffset = (UInt32)&param;
			proc = (VASearchProcPtr)&local_VADataSearchProc;
		}
		break;
	}
	nthItem = VASearch( proc, inOffset, inSearchData, outNext, inArray );
	return nthItem;
}

//===============================
VA_PROC UInt32 VAInternalBruteForceSearch( Size inSize, Boolean inSigned, UInt32 inOffset,
											void* inSearchData, VARef inArray )
{
	UInt32						nthItem = 0;
	VABruteForceSearchProcPtr	proc;
	
	switch (inSize)
	{
		case 1:
		proc = (inSigned) ? (VABruteForceSearchProcPtr)&local_VASignedByteSearchProc :
							(VABruteForceSearchProcPtr)&local_VAByteSearchProc;
		break;
		
		case 2:
		proc = (inSigned) ? (VABruteForceSearchProcPtr)&local_VASignedHalfWordSearchProc :
							(VABruteForceSearchProcPtr)&local_VAHalfWordSearchProc;
		break;
		
		case 4:
		proc = (inSigned) ? (VABruteForceSearchProcPtr)&local_VASignedWordSearchProc :
							(VABruteForceSearchProcPtr)&local_VAWordSearchProc;
		break;
		
		default:
		{
			VADataSortParam	param;
			
			param.size = inSize;
			param.offset = inOffset;
			inOffset = (UInt32)&param;
			proc = (VABruteForceSearchProcPtr)&local_VADataSearchProc;
		}
		break;
	}
	nthItem = VABruteForceSearch( proc, inOffset, inSearchData, inArray );
	return nthItem;
}

//===============================
static VAComparisonResult local_VAByteSearchProc( const Ptr inArrayDataP, UInt8* inSearchData,
												UInt32 inOffset )
{
	VAComparisonResult	result;
	UInt8*				p;
	
	p = (UInt8*)(inArrayDataP + inOffset);
	if (*inSearchData > *p) result = vaSearchDataIsGreater;
	else if (*inSearchData < *p) result = vaSearchDataIsLess;
	else result = vaEqual;
	return result;
}

//===============================
static VAComparisonResult local_VASignedByteSearchProc( const Ptr inArrayDataP,
														SInt8* inSearchData,
														UInt32 inOffset )
{
	VAComparisonResult	result;
	SInt8*				p;
	
	p = (SInt8*)(inArrayDataP + inOffset);
	if (*inSearchData > *p) result = vaSearchDataIsGreater;
	else if (*inSearchData < *p) result = vaSearchDataIsLess;
	else result = vaEqual;
	return result;
}

//===============================
static VAComparisonResult local_VAHalfWordSearchProc( const Ptr inArrayDataP,
													UInt16* inSearchData,
													UInt32 inOffset )
{
	VAComparisonResult	result;
	UInt16*				p;
	
	p = (UInt16*)(inArrayDataP + inOffset);
	if (*inSearchData > *p) result = vaSearchDataIsGreater;
	else if (*inSearchData < *p) result = vaSearchDataIsLess;
	else result = vaEqual;
	return result;
}

//===============================
static VAComparisonResult local_VASignedHalfWordSearchProc( const Ptr inArrayDataP,
															SInt16* inSearchData,
															UInt32 inOffset )
{
	VAComparisonResult	result;
	SInt16*				p;
	
	p = (SInt16*)(inArrayDataP + inOffset);
	if (*inSearchData > *p) result = vaSearchDataIsGreater;
	else if (*inSearchData < *p) result = vaSearchDataIsLess;
	else result = vaEqual;
	return result;
}

//===============================
static VAComparisonResult local_VAWordSearchProc( const Ptr inArrayDataP, UInt32* inSearchData,
													UInt32 inOffset )
{
	VAComparisonResult	result;
	UInt32*				p;
	
	p = (UInt32*)(inArrayDataP + inOffset);
	if (*inSearchData > *p) result = vaSearchDataIsGreater;
	else if (*inSearchData < *p) result = vaSearchDataIsLess;
	else result = vaEqual;
	return result;
}

//===============================
static VAComparisonResult local_VASignedWordSearchProc( const Ptr inArrayDataP,
														SInt32* inSearchData,
														UInt32 inOffset )
{
	VAComparisonResult	result;
	SInt32*				p;
	
	p = (SInt32*)(inArrayDataP + inOffset);
	if (*inSearchData > *p) result = vaSearchDataIsGreater;
	else if (*inSearchData < *p) result = vaSearchDataIsLess;
	else result = vaEqual;
	return result;
}

//===============================
static VAComparisonResult local_VADataSearchProc( const Ptr inArrayDataP,
												UInt8* inSearchData,
												VADataSortParam* inSortData )
{
	VAComparisonResult	result = vaEqual;
	Size				i;
	UInt8				*p;
	
	p = (UInt8*)(inArrayDataP + inSortData->offset);
	for (i = 1; i <= inSortData->size; i++, p++, inSearchData++)
	{
		if (*inSearchData > *p)
		{
			result = vaSearchDataIsGreater;
			break;
		}
		else if (*inSearchData < *p)
		{
			result = vaSearchDataIsLess;
			break;
		}
	}
	return result;
}

//===============================
VA_PROC void VANewNumberOnMovement( UInt32* ioNumber, UInt32 inCount, UInt32 inSource,
									UInt32 inDest )
{
	UInt32	number;
	
	number = *ioNumber;
	// See if number is one of the entries moved.
	if (number >= inSource && number < inSource + inCount)
	{
		number = inDest + number - inSource + 1;
		if (inSource <= inDest) number -= inCount;
	}
	else if (number > inSource && number <= inDest) number -= inCount;
	else if (number < inSource && number > inDest) number += inCount;
	*ioNumber = number;
}

//===============================
// It's safe to pass 0 for the number.
VA_PROC void VANewNumberOnAddition( UInt32* ioNumber, UInt32 inCount, UInt32 inAddHere )
{
	UInt32	number;
	
	number = *ioNumber;
	if (number >= inAddHere) number += inCount;
	*ioNumber = number;
}

//===============================
// It's safe to pass 0 for the number.
VA_PROC void VANewNumberOnDeletion( UInt32* ioNumber, UInt32 inCount, UInt32 inDeleteHere )
{
	UInt32	number;
	
	number = *ioNumber;
	if (number > inDeleteHere + inCount - 1) number -= inCount;
	else if (number >= inDeleteHere && number <= inDeleteHere + inCount - 1) number = 0;
	*ioNumber = number;
}

